! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module arrayutils

! Authors:  Lalit Jalota - XMM SOC
!          
! Description: Array utils by Lalit Jalota. Uses SLATEC
! 
! Comments:
!              
!  PV 02feb99 : - made into seperate module. Extracted from rgsframes v0.7
!  PV 18feb99 : - fixed special case of zero length array
!  PV 05mar99 : - added headArray. should move to saslib.
!      
! $Id: arrayutils.f90,v 1.2 1999/03/05 17:20:21 pvideler Exp $
!- 

  use saslib

  implicit none 
  ! ......................................................
  private
  public sortAndPurgeArray, headArray
  ! ......................................................

 interface headArray
    module procedure headArray_int8, headArray_int16, headArray_int32 
    module procedure headArray_real32, headArray_real64 
 end interface

contains  

!------------------------------------------------
  function headArray_int8(array) result(head)
!------------------------------------------------
 ! If size(array) > 10 return only the first 10 elements of array,
 ! otherwise return whole array. 
 ! PV 05mar99

  integer , parameter :: n=10
  integer (int8), dimension(:), intent(in)          :: array
  integer (int8), dimension( min(size(array),n))    :: head 

  head = array(1:min(size(array),n)) 

end function headArray_int8

!------------------------------------------------
  function headArray_int16(array) result(head)
!------------------------------------------------
  integer , parameter :: n=10
  integer (int16), dimension(:), intent(in)          :: array
  integer (int16), dimension( min(size(array),n))    :: head 

  head = array(1:min(size(array),n)) 

end function headArray_int16

!------------------------------------------------
  function headArray_int32(array) result(head)
!------------------------------------------------
 ! If size(array) > 10 return only the first 10 elements of array,
 ! otherwise return whole array. 
 ! PV 05mar99

  integer , parameter :: n=10
  integer (int32), dimension(:), intent(in)          :: array
  integer (int32), dimension( min(size(array),n))    :: head 

  head = array(1:min(size(array),n)) 

endfunction headArray_int32

!------------------------------------------------
  function headArray_real32(array) result(head)
!------------------------------------------------
  integer , parameter :: n=10
  real (single), dimension(:), intent(in)          :: array
  real (single), dimension( min(size(array),n))    :: head 
  head = array(1:min(size(array),n)) 
  end function 

!------------------------------------------------
  function headArray_real64(array) result(head)
!------------------------------------------------
  integer , parameter :: n=10
  real (double), dimension(:), intent(in)          :: array
  real (double), dimension( min(size(array),n))    :: head 
  head = array(1:min(size(array),n)) 
  end function 

  !====================================================================
  subroutine sortAndPurgeArray(rawArray, purgedArray, sizePurgedArray)
    !====================================================================

    ! this subroutine sorts an array using the SLATEC function qs2i1r, 
    ! then purges it, i.e. removes multiple occurences of same elements

    real, pointer, dimension(:)	    :: realArray
    integer(INT32),pointer, dimension(:) :: rawArray, purgedArray, &
         intArray
    integer			:: sizeRawArray, sizePurgedArray, i
    integer,parameter		:: sortOrder = 1

    sizeRawArray = size(rawArray)

    !!$ PV: qs2i1r does not like sorting a zero length array.
    !!      return zero length array purged array.
  
    if (sizeRawArray <= 0) then
       call message (appMsg, Verbose, 'sortAndPurgeArray: Size array <= 0')   
       allocate(purgedArray(0))
       sizePurgedArray = 0 
       RETURN      
    endif

    allocate(intArray(sizeRawArray))
    allocate(realArray(sizeRawArray))
    intArray = (/(i, i=0, sizeRawArray-1)/)
    realArray = real(intArray)

    ! sort array in ascending order
    call qs2i1r(rawArray, intArray, realArray, sizeRawArray, sortOrder)

    ! now purge array
    call purgeIntArray(rawArray, purgedArray, sizePurgedArray)

    return
  end subroutine sortAndPurgeArray

  !=================================================================
  subroutine purgeIntArray(inputArray, purgedArray, sizePurgedArray)
    !=================================================================

    ! this takes inputArray and returns purgedArray which is the same list
    ! except that multiple occurrences of elements are reduced to one.

    integer, dimension(:)		:: inputArray
    integer, pointer, dimension(:)	:: purgedArray
    integer 				:: sizeInputArray,sizePurgedArray,i,j

    sizeInputArray = size(inputArray)

    if (sizeInputArray > 0) then 
       allocate(purgedArray(0:sizeInputArray-1))
       purgedArray = 0
       purgedArray(0) = inputArray(1)
       j = 1
       do i=2,sizeInputArray
          if  (inputArray(i) /= inputArray(i-1)) then 
             purgedArray(j) = inputArray(i)
             j = j + 1
          end if
       end do
       sizePurgedArray = count(purgedArray /= 0)
    else
       write(errstr, *)" Input Array has zero length"
       call error("Error in purgeIntArray")
    end if

    return

  end subroutine purgeIntArray

  !=====================================================================
  subroutine matchIntArrays(array1, array2, matchedArray, nMatching)
    !====================================================================

    ! This subroutine takes two arrays and matches them sa much as possible
    ! and the matched array is returned as matchedArray, 
    ! The input arrays must be integer, sorted, and monotonic (no double
    ! values)
    !   PV19feb99: code by LJ. Not verified.

    integer, dimension(:)		:: array1, array2
    integer, pointer, dimension(:)	:: matchedArray
    integer				:: nMatching, i, j, k, 		&
         sizeArray1, sizeArray2, 	&
         lboundArray1, uboundArray1, 	&
         lboundArray2, uboundArray2, 	&
         biggerSize

    k = 0

    sizeArray1 = size(array1)
    sizeArray2 = size(array2)
    biggerSize = max(sizeArray1, sizeArray2)

    allocate(matchedArray(0:biggerSize-1))

    lboundArray1 = lbound(array1,1)
    uboundArray1 = ubound(array1,1)
    lboundArray2 = lbound(array2,1)
    uboundArray2 = ubound(array2,1)

    if (sizeArray1 >= sizeArray2) then

       FIRST:    do i = lboundArray2, uboundArray2-1	! smaller array

          SECOND:	     do j = i, uboundArray1		! larger array
             if (array1(j) >= array2(i+1))	then
                cycle FIRST
             end if

             if (array1(j) == array2(i) ) then
                matchedArray(k) = array2(i)
                k=k+1
                exit SECOND
             end if
          end do SECOND

       end do FIRST

    else 

       FIRST2:    do i = lboundArray1, uboundArray1-1	! smaller array

          SECOND2:      do j = i, uboundArray2		! larger array
             if (array2(j) >= array1(i+1))	then
                cycle FIRST2
             end if

             if (array2(j) == array1(i) ) then
                matchedArray(k) = array1(i)
                k=k+1
                exit SECOND2
             end if
          end do SECOND2

       end do FIRST2

    end if

    nMatching = k+1

    return

  end subroutine matchIntArrays

end module arrayutils

























