! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: rgsframes_mod
!
! Authors: Peter Videler XMM SOC, C. Erd <cerd@astro.estec.esa.nl>
!
! Description: Process RGS ODF file ready for further processing
!
! Comments: extensively re-write by CE 21-Mar-2001
!      21-05-2001 CE  added support for HTR mode
!      09-01-2002 JBa using reimplementation of duplicate finding
!
! $Id: rgsframes_mod.f90,v 1.92 2007/09/13 15:38:06 aibarra Exp $
!-

module rgsframes_mod

  use dal
  use cal
  use oal
  use param
  use saslib
  use sasdal
  use arrayutils
  use odfNames                  !in saslib/lib/saslib/odfnames.f90
  use eventAttributes

  implicit none
  private
  public rgsframes

  type(DataSetT) :: auxset
  public auxset
  type(tableT)  :: exposureTable, pixelsTable
  public exposureTable, pixelsTable

  logical notHTR
  public notHTR

  logical singleROnode
  public singleROnode

  real(double), dimension(:), allocatable :: alldeltaT !delta times for HTR mode
  real(double), dimension(9) :: avgDTime !averade delta times per CCD
  public alldeltaT, avgDTime

contains

  !------------------------------------------------------------
  subroutine rgsframes
    !------------------------------------------------------------
    use rgsframesutils
    use gti_mod

    integer :: datamode, dm, ccdnr, insubmode
    character(len=80) :: obsSubMode

    character(len=512)  :: filename
    character(len=600)  :: outfilename

    type(DataSetT) :: set
    type(tableT) :: gtiTable
    type(ColumnT) :: attributes_col, time_col, timedel_col
    integer, dimension(:), pointer :: attributes => null()

    type(ProposalInfoType) :: proposalInfo

    integer :: instrument, nexposures(2), exposureoffset, &
               exposure, nscif, i, j, l, ccdlo, ccdhi
    logical :: isAUXopen

    nexposures = 0
    exposureoffset = 0

    if ( .not. booleanParameter("processall") ) then
                                !Always copy pixel set (rgsframes is first in pipeline)
       call copyDataSet(stringParameter("set"), stringParameter("ccdset"))
       set = dataset(stringParameter("ccdset"), MODIFY)

       call OAL_setState(set)   !open set
       call OAL_getState(dataMode=datamode, ccdNr=ccdnr, &
                         exposureNr=exposureoffset) !get datamode
       exposureoffset = exposureoffset - 1
       if ( datamode == DATA_SPECTROSCOPY    ) nexposures(1) = 1
       if ( datamode == HIGH_TIME_RESOLUTION ) nexposures(2) = 1
    else
       call OAL_setState(instrument=instId(stringParameter("instrument")))
    end if

    j = 0
    isAUXopen = .false.
    LoopModes: do dm=DATA_SPECTROSCOPY, HIGH_TIME_RESOLUTION
       j = j +1
       if ( booleanParameter("processall") ) &
         nexposures(j) = OAL_selectScope(instId(stringParameter("instrument"))&
         , dataMode=dm)
       if ( nexposures(j) == 0 ) cycle LoopModes

       notHTR = dm == DATA_SPECTROSCOPY !set the HTR mode flag

       LoopExposures: do exposure=1,nexposures(j)
          nscif = OAL_SelectExposure( exposure + exposureoffset)
          if ( nscif == 0 ) cycle LoopExposures

          if ( booleanParameter("processall") ) then
             ccdlo = 1
             ccdhi = 9
          else
             ccdlo = ccdnr
             ccdhi = ccdnr
          end if

          LoopScienceFiles: do i=ccdlo,ccdhi
             if ( .not. OAL_hasSet(i) ) cycle LoopScienceFiles

             if ( booleanParameter("processall") ) then
!                if ( trim(stringParameter("instrument")) == 'rgs1' .and. i==7 ) then
!                   call warning('ImpossibleCCDID','RGS1-CCD7 does not exist!')
!                   cycle LoopScienceFiles
!                else if ( trim(stringParameter("instrument")) == 'rgs2' .and. i==4 ) then
!                   call warning('ImpossibleCCDID','RGS2-CCD4 does not exist!')
!                   cycle LoopScienceFiles
!                end if
                                !copy pixel sets
                filename = OAL_selectSet(i)
                l = len(trim(filename))
                write(outfilename,*) filename(l-30:l-7)//'_events.ds'
                call copyDataSet(filename, outfilename)
                set = dataset(outfilename, MODIFY)

                call OAL_setState(set)
             end if

                                !set DATAMODE attribute
             call setAttribute(set, 'DATAMODE'&
             , dataModeId(dm),comment="Instrument mode")
             call setAttribute(set, 'OBS_MODE', "POINTING"&
             , comment="Observation mode")

             call setAttribute(set,"FILTER", "UNKNOWN", comment="RGS has &
             &only one fixed filter")

             call OAL_proposalInfo(proposalInfo)
             call setAttribute(set,"OBJECT", trim(proposalInfo%TargetName)&
             , comment="Name of target")
             call setAttribute(set,"OBSERVER", trim(proposalInfo%GoName)&
             , comment="Name of GO")

             call CAL_setState(set) !set CAL state
      
             call CAL_getState(ccdModeId=insubmode)
             obsSubMode = modeId(insubmode)
             write(errstr,*) 'Observing mode was ',obsSubMode
             call message(AppMsg,Sparse)
             call setAttribute(set,"SUBMODE",obsSubMode, "Guest Observer mode")

             if ( .not. isAUXopen ) then !Get AUX file via OAL
                filename = OAL_selectAssociatedSet(AUXILIARY)
                auxSet = dataSet(filename, READ)
                isAUXopen = .true.

                                !and perform initial calculations
                if ( .not. notHTR ) call calculateHTRFrameTimes( set )
             end if
             
	     singleROnode = .false.
             
	     call addExposureTable(set)

             if ( notHTR ) then
                call processFrameTimes
             else
                call fillHTRFrameTimes
             end if

             call processPixelTable( set )

             call addFracExposure

             call Pixels_and_Frames

                                !GTI table
             attributes_col = column(exposureTable, EVATT_COLUMN_NAME, READ)
             time_col       = column(exposureTable, TIME_COLUMN_NAME , READ)
             timedel_col  = column(exposureTable, 'TIMEDEL'      , READ)

             attributes => int32Data(attributes_col)

             if ( verbosity(appMsg) >= Silent) then
                write(errstr,*) "Total number of frames flagged bad:"&
                , count(attributes /= 0), "out of a total of", size(attributes)
                call message(appMsg, Silent)
             end if

             if (all(attributes /= 0)) then
               write(errstr,'(a)') "All frames flagged as bad: &
               &therefore no GTI for this ccd."
               call warning('allFramesBad')
             end if

             gtitable = createGTI(set, good=(attributes == 0), &
                  time=real64Data(time_col), &
                  exposuretime=real32Data(timedel_col))

             call CAL_getState(instrument=instrument)
             call setAttribute(gtiTable, "INSTRUME", instId(instrument)&
             , comment="Instrument name")

             call setAttribute(gtiTable, "CCDOCB"&
             , int8Attribute(pixelsTable,"CCDOCB"), comment="On-chip binning")
             call setAttribute(gtiTable, "CCDID"&
             ,  int8Attribute(pixelsTable,"CCDID" ), comment="CCD identifier")

             if ( singleROnode ) then
	      call setAttribute(set,"SINGLERO", 1&
                 , comment="Is it Single RO node used?")
	     else
	      call setAttribute(set,"SINGLERO", 0&
                 , comment="Is it Single RO node used?")
	     end if

             call release( attributes_col )
             call release( time_col )
             call release( timedel_col )
             call release( gtiTable )

             call cleanupExposureTable

             call release(set)
          end do LoopScienceFiles

          if (isAUXopen) then
             call release( auxSet )
             isAUXopen = .false.
          end if
       end do LoopExposures
    end do LoopModes

  end subroutine rgsframes


  !---------------------------------------------------------------------
  subroutine processPixelTable (set)
  !---------------------------------------------------------------------
    use rgsframesutils

    type (ExposureInfoType) :: ExposureInfo  ! typedef from oal

    type(DataSetT) :: set
    type(ColumnT) :: frame_col, telx_col, tely_col, rawx_col, rawy_col
    type(ColumnT) :: ccdnode_col, ser_col, grade_col
    integer(kind=int8),  dimension(:), pointer :: ccdnode => null(), &
                                                  ser => null(), grade => null()
    integer(kind=int16), dimension(:), pointer :: telx => null(), &
                                                  tely => null(), rawx => null(), rawy => null()
    integer(kind=int32), dimension(:), pointer :: frame => null()

    character(len=4) :: dppVersion
    integer :: instrument, i, ccdmode
    integer :: windowx0, windowdx, windowy0, windowdy, binning
    integer :: maxdx, maxdy, nflips, tobedenoded
    real(double) :: telapse
    type(IntStack) :: duplicates
    
    tobedenoded = 0

    call CAL_getState(instrument=instrument)

    select case (instId(instrument))   ! table names depend on instrument
    case ('RGS1')
       if ( notHTR ) then
          pixelsTable = table(set, 'R1SPE1')
       else
          pixelsTable = table(set, 'R1HTE1')
       end if
       call rename(pixelsTable ,'PIXELS')
    case ('RGS2')
       if ( notHTR ) then
          pixelsTable = table(set, 'R2SPE1')
       else
          pixelsTable = table(set, 'R2HTE1')
       end if
       call rename(pixelsTable ,'PIXELS')
    case default
       call error ('instIdNotFoundPixels','instrument string '//instId(instrument))
    end select


    !------------------------
    ! add standard attributes
    !------------------------
    call CAL_getState(ccdModeId=ccdmode)
    call setAttribute(pixelsTable,"CCDMODE",modeId(ccdmode), "CCD mode")

    !set Elapsed time
    call OAL_exposureInfo(exposureInfo)

    telapse = ToTimeTag(exposureInfo%actualEndTime) &
            - ToTimeTag(exposureInfo%actualStartTime)
    call setAttribute(pixelsTable, "TSTART", ToTimeTag(exposureInfo%actualStartTime), &
    	units="s", comment="Exposure Start Time")
    call setAttribute(pixelsTable, "TSTOP", ToTimeTag(exposureInfo%actualEndTime), &
    	units="s", comment="Exposure End Time")
    call setAttribute(pixelsTable, "TELAPSE", telapse,units="s", &
        comment="Time elapsed between start and end of observation")
    call setAttribute(pixelsTable, "TIMESYS", "TT", comment="XMM time is TT (Terrestial Time)")
    call setAttribute(pixelsTable, "MJDREF", 50814, comment="1998-01-01T00:00:00 (TT) expressed in MJD (TT)")
    call setAttribute(pixelsTable, "TIMEZERO", 0, units="s", comment="Clock correction (if not zero)")
    call setAttribute(pixelsTable, "TIMEUNIT","s", comment="All times in s unless specified otherwise")
    call setAttribute(pixelsTable, "CLOCKAPP", .true., comment="Clock correction applied")
    call setAttribute(pixelsTable, "TIMEREF", "LOCAL", comment="Reference location of photon arrival times")
    call setAttribute(pixelsTable, "TASSIGN", "SATELLITE", comment="Location of time assignment")

                                !delete unwanted set attributes
    if (hasAttribute(set, "CATEGORY")) call deleteAttribute(set, "CATEGORY")
    if (hasAttribute(set, "CREATOR"))  call deleteAttribute(set, "CREATOR")

                                !copy some attributes from pixel table to set
    call copySomeAttributes(attributable(pixelsTable), attributable(set))


    !----------------------------------------------
    ! correct for int16 wrap around of FRAME number
    !----------------------------------------------

    frame_col = column(pixelsTable, FRAME_COLUMN_NAME, MODIFY)
    frame => int32Data(frame_col)

                                !Patch for wrap around of frame numbers in CCD FRAME column
    call CorrectWrapAround(frame, nflips)
    if (nflips > 0) then
       call warning('wrapAround',' Frame numbers corrected for int16 wrap around &
                    &in table '//trim(qualifiedName(pixelsTable)) //' .')
    endif


    !-----------------------------------------
    ! rename telemetry RAWX/Y colums to TELX/Y
    !-----------------------------------------

    telx_col = column(pixelsTable, 'RAWX', MODIFY)
    call rename(telx_col, 'TELX')
    if ( notHTR ) then
       tely_col = column(pixelsTable, 'RAWY', MODIFY)
       call rename(tely_col, 'TELY')

    else
       call deleteColumn(pixelsTable, 'RAWY')
    end if


    !--------------------
    ! sort PIXELS table
    !--------------------
                                !sort PIXELS table on frame number
    call message(appMsg, Verbose, "Evaluating monotonicity of PIXELS table ...")
    if (.not. monotonicIncreasing(frame) ) then
       call message(appMsg, Verbose,"Sorting EXPOSURE table")
       call tableSort(pixelsTable, frame_col, sortorder=1)   ! from saslib/lib/sasdal
    end if


    !-------------------------------
    ! find and delete duplicate rows
    !-------------------------------

    telx => int16Data(telx_col)
    if ( notHTR ) tely => int16Data(tely_col)

    ccdnode_col = column(pixelsTable, 'CCDNODE', MODIFY)
    ccdnode => int8Data(ccdnode_col)

    write(errstr,*) 'Looking for duplications in PIXELS table with', &
         size(frame), 'pixels ...'
    call message(appMsg,Verbose)

    ! find duplicate entries
    call construct(duplicates)
    if ( notHTR ) then
       duplicates = findDuplicates(frame,int(telx),int(tely),int(ccdnode))
    else
       duplicates = findDuplicates(frame,int(telx),int(ccdnode)) ! do not need tely
    end if

    call release( frame_col )
    call release( telx_col )
    if ( notHTR ) call release( tely_col )
!    call release( ccdnode_col )   / Releasing it later

    ! remove duplicate entries
    write (errstr,*) 'Identified',size(duplicates),&
         'duplicate entries in PIXELS table'
    call message(appMsg,Silent)
    if (.not.empty(duplicates)) then
       do while (.not.empty(duplicates))
          call deleteRows( pixelsTable, top(duplicates)-1, 1)
          call pop(duplicates)
       end do
       write (errstr,*) 'Removed duplicate entries in PIXELS table'
       call message(appMsg,Silent)
! AI. 13/09/2007
! AI. Releasing it now (duplicate entries)
! AI. If there are duplicate entries the pixelTable is resized and
! AI. therefore ccdnode array has to be also resized.
! AI. To match the rawx size.
       call release( ccdnode_col )   
       ccdnode_col = column(pixelsTable, 'CCDNODE', MODIFY)
       ccdnode => int8Data(ccdnode_col)
!AI: Single read out node change 
    endif
    call destruct(duplicates)
                                !get window details


    binning  = int16Attribute(pixelsTable,"CCDOCB")
    windowx0 = int16Attribute(pixelsTable,"WINDOWX0")
    windowy0 = int16Attribute(pixelsTable,"WINDOWY0")
    windowdx = int16Attribute(pixelsTable,"WINDOWDX")
    windowdy = int16Attribute(pixelsTable,"WINDOWDY")


                                !get size of chip in binned pixel sizes

    select case (binning)
    case(1)
       maxdx = 1024             ! PV these values are hardcoded. ###
       maxdy = 384
    case (3)
       maxdx = 342
       maxdy = 128
       if (windowdx == 342) then
         singleROnode = .true.
         tobedenoded = 1
         call setAttribute(pixelsTable, "WINDOWDX",171, comment="CCD-n X Length - Changed for single readout node")
         call setAttribute(pixelsTable, "ORIG_WDX",342, comment="Real window DX - Changed for single readout node")
       end if 
    case (5)
       maxdx = 205
       maxdy =  77
    end select

                                !we also need to check window sizes for given OCB
    if ( notHTR ) then
       if ( windowx0+windowdx > maxdx .or. windowy0+windowdy > maxdy ) then
          write(errstr,*) "inconsistent window values"
          call error("windowSize")
       end if
    else
       if ( windowx0+windowdx > maxdx ) then
          write(errstr,*) "inconsistent window values"
          call error("windowSize")
       end if
    end if


    !-----------------------------------------------
    ! create new RAWX/Y and shift pixel number by +1
    !-----------------------------------------------

                                !update window0 keyword, shift by +1
    call setAttribute(pixelsTable,"WINDOWX0",windowx0+1,units='pixel')
    if ( notHTR ) call setAttribute(pixelsTable,"WINDOWY0",windowy0+1,units='pixel')

                                !make new columns
    telx_col = column(pixelsTable, 'TELX', READ)
    telx => int16Data(telx_col)
    rawx_col = addColumn(pixelsTable,"RAWX", integer16, units="pixel", comment="standard definition of RAWX")
    rawx => int16Data(rawx_col)
    rawx = telx + 1


    !-------------------------------
    ! if single readout node change RAWX and node 
    !-------------------------------
    
    if (tobedenoded ==1) then
	where (rawx > 171) 
		rawx = (343 - rawx) 
		telx = (341 - telx)
		ccdnode = 1
	end where	     
    end if
        

    call release( ccdnode_col )  
    call release ( telx_col )
    call release ( rawx_col )

    if ( notHTR ) then
       tely_col = column(pixelsTable, 'TELY', READ)
       tely => int16Data(tely_col)
       rawy_col = addColumn(pixelsTable,"RAWY", integer16, units="pixel", comment="standard definition of RAWY")
       rawy => int16Data(rawy_col)
       rawy = tely + 1

       call release ( tely_col )
       call release ( rawy_col )
    end if

    !-------------------------------
    ! replace SER by GRADE and SHAPE
    !-------------------------------

    if ( notHTR ) then

    ser_col = column(pixelsTable, 'SER', READ)
    ser =>  int8Data(ser_col)
    call rename(ser_col, 'SHAPE')
    grade_col = addColumn(pixelsTable,'GRADE', integer8, comment="Event grade number")
    grade => int8Data(grade_col)

!!$       From email Doug dd 01oct99
!!$
!!$          DPP version 16      DPP version 16.2
!!$        -----------------    -----------------
!!$        SER  SHAPE  GRADE    SER  SHAPE  GRADE
!!$         0     0      1       0     0      1
!!$         1    null    2       1     1      4
!!$         2    null    3       2     2      2
!!$         3     1      4       3     3      2
!!$                              4     4      3
!!$                              5     5      3
!!$                              6     6      3
!!$                              7     7      3


    if (any(ser<0)) call error ('wrongSER1','SER value <0 found. &
         & SER must be in range [0-3] for version v16 of DPP software.')

    ! dpp version string is hardcoded to format 'xx.x'
    write (dppVersion,'(F4.1)') CAL_getMiscellaneousDataValue("DPPVERS")
    call message(appMsg, Verbose, 'DPP version = '//dppVersion )

    !       select case (stringParameter('dppversion'))
    select case (dppVersion)

    case ('0.0')
       call warning('grade1ser0','dppVersion=0.0 Setting grade=1 ser=0 for all events')
       grade=1     ! for testing only
       ser=0
    case('16.0')
       if (any(ser>3)) call error ('wrongSER2','SER value >3 found. &
            & SER must be in range [0-3] for version v16 of DPP software.')
       grade = ser + 1

       if (.not. nullDefined(ser_col)) call setNullValue(ser_col, 255)

       do i=0, size(ser) - 1
          if (ser(i)==1 .or. ser(i)==2) call setToNull(ser_col ,i)
          if (ser(i)==3) ser(i)=1
       enddo
    case ('16.2', '19.0')
       grade = 3
       where(ser == 0) grade=1
       where(ser == 1) grade=4
       where(ser==2 .or. ser==3) grade=2
    case default
       call error ('dppIdNotFound','DPP version string')
    end select

    call release( ser_col )
    call release( grade_col )

    end if                      !end notHTR      
      
  end subroutine processPixelTable


  !---------------------------------------------------------------------
  subroutine Pixels_and_Frames
  !---------------------------------------------------------------------
    use rgsframesutils

    type(ColumnT) :: exposureframe_col, pixelsframe_col, pixelrate_col, flags_col
    type(ColumnT) :: timedel_col
    integer, dimension(:), pointer :: exposureframe => null(), pixelsframe => null(),&
                                      flags => null(),  pixelrate => null()
    real   , dimension(:), pointer :: timedel => null()

    integer, dimension(:), allocatable :: pixelsPerFr
    integer :: i, n

                                !at the moment only RATE related stuff
    if ( .not. booleanParameter("withmaxrawpixels") .and. &
         .not. booleanParameter("withmaxrawpixelrate") .and. &
         .not. booleanParameter("withratecolumn")) return

    flags_col = Column(exposureTable, EVATT_COLUMN_NAME, MODIFY)
    flags => int32Data(flags_col)

    exposureframe_col = column(exposureTable, FRAME_COLUMN_NAME, READ)
    pixelsframe_col = column(pixelsTable, FRAME_COLUMN_NAME, READ)
    exposureframe => int32Data(exposureframe_col)
    pixelsframe => int32Data(pixelsframe_col)

    allocate(pixelsPerFr(lbound(exposureframe,1):ubound(exposureframe,1)))
    pixelsPerFr = pixelsPerFrame(pixelFrame=pixelsframe , frame=exposureframe)

    call release( pixelsframe_col )

                                !add a rate column
    if ( booleanParameter("withratecolumn") ) then
       pixelrate_col = addColumn(exposureTable,'PIXELRATE', &
            integer32,units="count /frame",comment="Number of raw &
            &pixels per frame")
       pixelrate => int32Data(pixelrate_col)
       pixelrate = pixelsPerFr

       call release( pixelrate_col )
    end if

                                !limit number of pixels per frame
    if ( booleanParameter("withmaxrawpixels") ) then
       where ( pixelsPerFr > intParameter("maxrawpixels") )&
         flags = EVATT_IN_BAD_FRAME

       if ( verbosity(appMsg) >= Silent) then
          write(errstr,*) count( pixelsPerFr > intParameter("maxrawpixels") ), &
               "frames flagged bad based on maxrawpixels"
          call message(appMsg,Silent)
       end if

       if (all(pixelsPerFr > intParameter("maxrawpixels"))) then
         write(errstr,'(a)') "All frames flagged as EVATT_IN_BAD_FRAME &
         &because the number of pixels per frame is greater than &
         &--maxrawpixels: therefore no GTI for this ccd."
         call warning('allFramesTooManyPixels')
       end if
    end if

                                !limit rate of pixels
    if ( booleanParameter("withmaxrawpixelrate") ) then

       timedel_col = column(exposureTable,'TIMEDEL',READ)
       timedel => real32Data(timedel_col)

       n = 0
       do i=0, ubound(exposureframe,1)
         if (timedel(i) > 0) then
           if (pixelsPerFr(i)/timedel(i)>realParameter("maxrawpixelrate")) then
              flags(i) = EVATT_IN_BAD_FRAME
              n = n +1
           end if
         end if
       end do
       write(errstr,*) n, "frames flagged bad based on maxrawpixelrate"
       call message(appMsg,Silent)

       if (n .ge. size(flags)) then
         write(errstr,'(a)') "All frames flagged as EVATT_IN_BAD_FRAME &
         &because the pixel rate is greater than --maxrawpixelrate: &
         &therefore no GTI for this ccd."
         call warning('allFramesTooRapidPixels')
       end if
    end if

    call release( flags_col )
    call release( exposureframe_col )
    deallocate(pixelsPerFr)
  end subroutine Pixels_and_Frames


  !---------------------------------------------------------------------
  subroutine addExposureTable (set)
  !---------------------------------------------------------------------
    use rgsframesutils

    type(DataSetT) :: set
    type(TableT) :: auxTable, tmptable    !the AUX table
    type(ColumnT) :: ccdid_col, col, eosfine_all_col
    integer(kind=int16), dimension(:), pointer :: i16 => null()
    integer(kind=int32), dimension(:), pointer :: eosfine_all => null()
    integer, dimension(:), allocatable :: auxindex
    integer :: nrows

                                !EXPOSURE table
    type(ColumnT) :: frame_col, seqindex_col, eoscoars_col, eosfine_col
    type(ColumnT) :: nlostevt_col, abortflg_col, ndpp_col, flag_col
    integer(kind=int32), dimension(:), pointer :: frame => null(), eoscoars => null(),&
                                                  eosfine => null(), ndpp => null(), flag => null()
    integer(kind=int16), dimension(:), pointer :: nlostevt => null(), abortflg => null()
    integer(kind=int8),  dimension(:), pointer :: ccdid => null(), seqindex => null()

    type(tableT) :: auxTable2
    type(ColumnT) :: windowdx_col, windowdy_col
    integer(kind=int16), dimension(:), pointer :: windowdx => null(), windowdy => null()

    integer :: theccdid, instrument
    integer :: i, j
    type(IntStack) :: duplicates

    integer(kind=int32) :: nodeAreaInPixels, row

    !--------------------------------
    ! copy AUX1 table to EXPOSURE table
    !--------------------------------

    call CAL_getState(instrument=instrument, ccdChipId=theccdid)
    write(errstr,*) "Starting addExposureTable, CCDID=", theccdid
    call message(appMsg, Verbose)

    select case (instId(instrument)) !table names depend on instrument
    case ('RGS1')
       auxTable = table(auxSet, 'R1AUX1', READ)
    case ('RGS2')
       auxTable = table(auxSet, 'R2AUX1', READ)
    case default
       call error ('instIdNotFoundExposure'&
         ,'instrument string '//instId(instrument))
    end select
                                !find rows for this CCD
    ccdid_col = column(auxTable, CCDID_COLUMN_NAME, READ)
    ccdid    => int8data(ccdid_col)

    nrows = count( ccdid == theccdid )

    allocate(auxindex(nrows))
    j = 0
    do i=0,size(ccdid)-1                !find rows that have the current CCDID
       if ( ccdid(i) == theccdid ) then
          j = j+1
          auxindex(j) = i
       end if
    end do

    call release( ccdid_col )

    exposureTable = addTable(set, 'EXPOSURE', nrows, "Frame and Exposure &
        &information")
    call setAttribute(exposureTable,"INSTRUME", instId(instrument)&
        , comment="Instrument name")
    call setAttribute(exposureTable,"CCDID", theccdid, comment="CCD identifyer")

                                !add and copy columns
    call selectedAddnCopyColumn( auxTable, FRAME_COLUMN_NAME, auxindex, &
         exposureTable, FRAME_COLUMN_NAME, 'counter','Frame number')

                                !--------------------
                                !limited copy for HTR
                                !--------------------
    if ( notHTR ) then
                                !need to copy from int16 to int8
       col = column(auxTable, SEQINDEX_COLUMN_NAME, READ)
       i16 => int16Data(col)
       seqindex_col = addColumn(exposureTable, SEQINDEX_COLUMN_NAME, integer8, &
            comment='Readout sequence index' )
       seqindex => int8Data(seqindex_col)
       seqindex = i16(auxindex)
       call release( col )
       call release( seqindex_col )

       call selectedAddnCopyColumn( auxTable, "NUPPERC", auxindex, &
            exposureTable, "REJPIX0", 'count', 'Number of pixels above &
            &upper thrshold' )

       call selectedAddnCopyColumn( auxTable, "NUPPERD", auxindex, &
            exposureTable, "REJPIX1", 'count', 'Number of pixels above &
            &upper thrshold' )

                       !below are temporary columns needed for local processing
       call selectedAddnCopyColumn( auxTable, "EOSCOARS", auxindex, &
            exposureTable, "EOSCOARS", 's' )

!    call selectedAddnCopyColumn( auxTable, "EOSFINE", auxindex, &
!         exposureTable, "EOSFINESHORT",  uinteger16 )

       call selectedAddnCopyColumn( auxTable, "NDPP", auxindex, &
            exposureTable, "NDPP" )

                              !need to convert EOSFINE from unsigned16 to int32
       col = Column(auxTable, "EOSFINE", READ)
       tmptable = addTable(set, "TMPTABLE", numberOfRows(auxTable))
       eosfine_all_col = addColumn(tmptable,"EOSFINE", integer32)
       call setData( eosfine_all_col, col ) !ocpy unsigned to signed
       eosfine_all => int32Data( eosfine_all_col )

       eosfine_col = addColumn(exposureTable, "EOSFINE", integer32)
       eosfine => int32Data( eosfine_col )
       eosfine = eosfine_all( auxindex )

       call release( col )
       call release( eosfine_all_col )
       call release( tmptable )
       call deleteBlock(set, "TMPTABLE" )
       call release( eosfine_col )
    end if                      !notHTR

    call selectedAddnCopyColumn( auxTable, "NLOSTEVT", auxindex, &
         exposureTable, "NLOSTEVT" )

    call selectedAddnCopyColumn( auxTable, "ABORTFLG", auxindex, &
         exposureTable, "ABORTFLG" )

    call release( auxTable )


    !--------------------
    ! sort EXPOSURE table
    !--------------------

    frame_col = column(exposureTable, FRAME_COLUMN_NAME, MODIFY)
    frame => int32Data( frame_col )

                                !sort AUX table on frame number
    call message(appMsg, Verbose, "Evaluating monotonicity of EXPOSURE &
    &table ...")
    if (.not. monotonicIncreasing(frame) ) then
       call message(appMsg, Verbose,"Sorting EXPOSURE table")
       call tableSort(exposureTable, frame_col, sortorder=1)
       ! from saslib/lib/sasdal
    end if

    !-------------------------------------------------
    ! find and delete duplicate rows of EXPOSURE table
    !-------------------------------------------------

    if ( notHTR ) then
       eoscoars_col = column(exposureTable, "EOSCOARS", MODIFY)
       eoscoars => int32Data( eoscoars_col )
    end if

    write(errstr,*) 'Looking for duplications in EXPOSURE table with', &
         size(frame), 'frames ...'
    call message(appMsg, Verbose)

    ! find duplicate entries
    call construct(duplicates)
    if ( notHTR ) then
       duplicates = findDuplicates(frame,eoscoars)
       call release( eoscoars_col )
    else
       duplicates = findDuplicates(frame)
    end if
    call release( frame_col )

    ! remove duplicate entries
    write (errstr,*) 'Identified',size(duplicates),&
         'duplicate entries in EXPOSURE table'
    call message(appMsg,Silent)
    if (.not.empty(duplicates)) then
       do while (.not.empty(duplicates))
          call deleteRows( exposureTable, top(duplicates)-1, 1)
          call pop(duplicates)
       end do
       write (errstr,*) 'Removed duplicate entries in EXPOSURE table'
       call message(appMsg,Silent)
    end if
    call destruct(duplicates)

    !add flag column
    flag_col = addColumn(exposureTable, EVATT_COLUMN_NAME, integer32&
    , comment='Frame attributes')
    flag => int32Data(flag_col)
    call setEventAttribute(exposureTable, EVATT_IN_BAD_FRAME)


    !---------------------
    ! check EXPOSURE table
    !---------------------

    nlostevt_col = column(exposureTable,"NLOSTEVT" ,READ)
    abortflg_col = column(exposureTable,"ABORTFLG" ,READ)
    nlostevt => int16Data(nlostevt_col)
    abortflg => int16Data(abortflg_col)
    if ( notHTR ) then
       ndpp_col     = column(exposureTable,"NDPP"     ,READ)
       ndpp     => int32Data(ndpp_col)
    end if

                           !get WINDOW parameters from second table in AUX file
    select case (instId(instrument)) !table names depend on instrument
    case ('RGS1')
       auxTable2 = table(auxSet, 'R1AUX2', READ)
    case ('RGS2')
       auxTable2 = table(auxSet, 'R2AUX2', READ)
    end select
    ccdid_col    = column(auxTable2, 'CCDID'   , READ)
    windowdx_col = column(auxTable2, 'WINDOWDX', READ)
    windowdy_col = column(auxTable2, 'WINDOWDY', READ)
    ccdid => int8Data(ccdid_col)
    windowdx => int16Data(windowdx_col)
    windowdy => int16Data(windowdy_col)

    do i=0,ubound(windowdx,1)
      if ( ccdid(i) == theccdid ) then
        row = i
    exit
      end if
    end do
    if ( row > ubound(windowdx,1) ) then
       write(errstr,*) 'Could not find CCDID=',theccdid,'in AUX2.'
       call error('missingCCDinAUX2')
    end if

    nodeAreaInPixels = int(windowdx(row), int32) * windowdy(row)

    if (all(abortflg /= 2)) then
      write(errstr,'(a, i2.2)') "No frame in the exposure table has &
      &ABORTFLG==2: therefore no GTI for ccd ", theccdid
      call warning('badAbortFlagAllFrames')
    end if

    if ( notHTR ) then
                                !hardcoded condition nlostevt=0 abortflg=2
      if (all(nlostevt /= 0)) then
        write(errstr,'(a, i2.2)') "All frames in the exposure table have lost &
        &events - therefore no GTI for ccd ", theccdid
        call warning('eventsLostAllFrames')
      end if
      if (all(ndpp /= nodeAreaInPixels)) then
        write(errstr,'(a, i2.2, a, i2.2)') "All frames in table R?AUX1 of &
        &the aux file have NDPP values different from WINDOWDX*WINDOWDY &
        &(for row ", row+1, ") in table R?AUX2 of the same file. Therefore &
        &no GTI for ccd ", theccdid
        call warning('badNdppAllFrames')
      end if

      where ( nlostevt /= 0 .or. abortflg /= 2&
      .or. ndpp /= nodeAreaInPixels )
        flag = EVATT_IN_BAD_FRAME
      elsewhere
        flag = 0
      end where
    else
                                !hardcoded condition nlostevt=2 abortflg=2

      if (all(nlostevt /= 2)) then
        write(errstr,'(a, i2.2)') "All frames in the exposure table have !=2 &
        &lost events - therefore no GTI for ccd ", theccdid
        call warning('htrEventsLostAllFrames')
      end if

      where( nlostevt /= 2 .or. abortflg /= 2 )
        flag = EVATT_IN_BAD_FRAME
      elsewhere
        flag = 0
      end where
    end if

    flag(0) = EVATT_IN_BAD_FRAME !first frame is bad, too

!    if ( verbosity(appMsg) >= Silent) then
       write(errstr,*) count( flag /= 0 ), "frames flagged bad based on instrument bad-condition flags"
       call message(appMsg,Silent)
!    end if

!    if (all(flag == EVATT_IN_BAD_FRAME)) then
!      write(errstr,'(a, i2.2)') "All frames flagged as EVATT_IN_BAD_FRAME: &
!      &therefore no GTI for ccd ", theccdid
!      call warning('allFramesBadEvatt')
!    end if

    call release( nlostevt_col ) !from EXPOSURE table
    call release( abortflg_col )
    if ( notHTR ) call release( ndpp_col )
    call release( flag_col )

    call release( ccdid_col )   !from AUX2 table
    call release( windowdx_col )
    call release( windowdy_col )
    call release( auxTable2 )

    call message(appMsg, Verbose, "Finished addExposureTable")
  end subroutine addExposureTable


  !---------------------------------------------------------------------
  subroutine processFrameTimes
  !---------------------------------------------------------------------

    type(ColumnT) :: frame_col, eoscoars_col, eosfine_col, seqindex_col
    type(ColumnT) :: timedel_col, time_col, flag_col

    integer(kind=int32), dimension(:), pointer :: frame => null(), &
                                                  eoscoars => null(), &
                                                  eosfine => null(), flag => null()
    integer(kind=int8), dimension(:), pointer :: seqindex  => null()
    real, dimension(:), pointer :: timedel => null()
    real(double), dimension(:), pointer :: time => null(), obtTag => null()

    type(tableT) :: auxTable2
    type(ColumnT) :: seqCCDIDList_col
    integer(kind=int8), dimension(:), pointer :: seqCCDIDList => null()

    integer :: instrument, theccdid, seqLength, i, nreads
    real(double) :: expStartTime
    integer, dimension(:), allocatable :: delta_frames, slots
    real :: avgfrtime

    if ( numberOfRows(exposureTable) < 1 ) RETURN

    frame_col    = column(exposureTable, FRAME_COLUMN_NAME, READ)
    frame    => int32Data( frame_col )

    seqindex_col = column(exposureTable, SEQINDEX_COLUMN_NAME , READ)
    seqindex => int8Data( seqindex_col )
!    if (seqindex(1) == 0) call error("frameExposureTimes:HTR", &
!              "First value of SEQINDEX column is zero. &
!             & This probably resulted from an HTR mode file (*HTE.FIT). &
!             & This routine can not be used to calculate exposuretimes &
!             & for High Time Resolution mode observations.")


    !----------------------------------
    ! calculate frame integration times
    !----------------------------------

    eoscoars_col = column(exposureTable, 'EOSCOARS', READ)
    eosfine_col  = column(exposureTable, 'EOSFINE' , READ)
    eoscoars => int32Data( eoscoars_col )
    eosfine  => int32Data( eosfine_col )

    call CAL_getState(instrument=instrument, ccdChipId=theccdid, &
         exposureStartTime=expStartTime )
    select case (instId(instrument)) !table names depend on instrument
    case ('RGS1')
       auxTable2 = table(auxSet, 'R1AUX2', READ)
    case ('RGS2')
       auxTable2 = table(auxSet, 'R2AUX2', READ)
    end select

                                !get the readout sequence from AUX2
    seqCCDIDList_col = column(auxTable2, 'CCDID', READ)
    seqCCDIDList => int8Data (seqCCDIDList_col)
    seqLength = count(seqCCDIDList /= 0) !total length of readout sequence

    allocate(slots(seqLength))
    slots = 0
    nreads = count( seqCCDIDList == theccdid )
                                !readout slot numbers for current CCD
    slots(1:nreads) = pack( (/ (i,i=1,size(seqCCDIDList)) /), seqCCDIDList == theccdid )

    allocate(delta_frames(seqLength))
    delta_frames = 0

    SeqTable: do i=1,nreads               !fill number of frames from previous readout
       if ( i == 1 ) then
          delta_frames(slots(i)) = slots(i)
       else
          delta_frames(slots(i)) = slots(i) - slots(i-1)
       end if
    end do SeqTable
    delta_frames(slots(1)) = seqLength - slots(nreads) +slots(1)

                                !add frame integration times
    timedel_col = addColumn(exposureTable, 'TIMEDEL', real32, units='s', comment='Frame integration time')
    timedel => real32Data(timedel_col)
    timedel = 0

                                !fill frametime by find time elapsed to previous frame
    Frames: do i=lbound(frame,1)+1,ubound(frame,1)

                                !seq index must match ccdnr
       if ( seqCCDIDList(seqindex(i)-1) /= theccdid ) then
             write(errstr,*) 'Sequence index and CCDid do not match for &
                  & frame number ', frame(i)
             call error('sequenceIndexMismatch')
       endif

                                !frame numbers must match seqence length. Otherwise frames are missing
       if ( frame(i) /= frame(i-1)+delta_frames(seqindex(i)) ) then
          write(errstr,*) 'Missing frame detected in EXPOSURE table before frame number',frame(i),&
               '; Estimating frame time from readout sequence.'
           call message(appMsg, Sparse)

        else
                                !eoscoarse is in sec. eosfine in units of 2**-16 sec (2**16=
           timedel(i) = eoscoars(i) - eoscoars(i-1) + &
               ( eosfine(i) - eosfine(i-1) ) / 65536.
        end if
    end do Frames

                                !estimate average time to read one CCD
    avgFrtime = sum( pack(timedel/delta_frames(seqindex), timedel>0) ) / count(timedel > 1e-9)
    where ( timedel < 1e-9 ) timedel = avgFrTime * delta_frames(seqindex)

    deallocate(slots,delta_frames)

    !---------------------------
    ! calculate frame event time
    !---------------------------
                                !add TIME column
    ! see RGS telemetry document RGS-MSSL-SW-002 (v7.0 24aug98 page 7)
    ! timestamp from OAL is END of frame integration time:
    time_col = addColumn(exposureTable, TIME_COLUMN_NAME, real64, units='s'&
        , comment='seconds since MJDREF')
    time => real64Data(time_col)

                                !get time from OAL
    call OAL_frameCounterToObt(frameCounter=frame, obt=obtTag)

                                !Patch for odffix-ed data not having proper
                                !Time correlation file

                                !this is no longer necessary with DAL 3.67
!    if (booleanParameter("timepatch")) then
!       write(errstr,*) "Subtracting offset from Obt tags. Value : ", eoscoars(0)
!       call warning('timePatch')
!       obtTag = obtTag - eoscoars(0)
!    end if

    call OAL_obtToTimeTag(obt=obtTag, time=time)
    call OAL_releaseMemory(obtTag)

    time = time - timedel/2. !subtract half of exposure time to get mid of frame

                                !check lower limit on frame times
    flag_col = Column(exposureTable, EVATT_COLUMN_NAME)
    flag => int32Data(flag_col)
    where ( timedel < realParameter("minexposuretime") )&
      flag = EVATT_IN_BAD_FRAME

    if ( verbosity(appMsg) >= Silent) then
       write(errstr,*) count( timedel < realParameter("minexposuretime") ), &
            "frames flagged bad based on minexposuretime"
       call message(appMsg,Silent)
    end if

    if (all(timedel < realParameter("minexposuretime"))) then
      write(errstr,'(a, i2.2)') "All frames flagged as EVATT_IN_BAD_FRAME &
      &because TIMEDEL shorter than --minexposuretime: &
      &therefore no GTI for ccd ", theccdid
      call warning('allFramesTooShort')
    end if

    call release( timedel_col )
    call release( time_col )
    call release( flag_col )
    call release( eoscoars_col )
    call release( eosfine_col )

    call release(seqCCDIDList_col)
    call release(auxTable2)

  end subroutine processFrameTimes


  !---------------------------------------------------------------------
  subroutine addFracExposure
  !---------------------------------------------------------------------
    !add exposure fractions due to on-board rejected pixels

                                !EXPOSURE table
    type(ColumnT) :: rejpix0_col, rejpix1_col, fracexp0_col, fracexp1_col
    integer, dimension(:), pointer :: rejpix0 => null(), rejpix1 => null()
    real, dimension(:), pointer :: fracexp0 => null(), fracexp1 => null()

                                !interface with CAL
    integer(kind=int16), dimension(:), pointer  :: rawX => null(), rawY => null(), badtype => null(), yextent => null()
    logical, dimension(:), pointer              :: uplinked => null()

    integer :: inode, i, nuplinked(2)=0, allbins

    fracexp0_col = addColumn(exposureTable, "FRACEXP0", real32, comment='Exposure fraction node 0')
    fracexp1_col = addColumn(exposureTable, "FRACEXP1", real32, comment='Exposure fraction node 1')
    fracexp0 => real32Data( fracexp0_col )
    fracexp1 => real32Data( fracexp1_col )

    if ( notHTR ) then
       rejpix0_col = column(exposureTable, "REJPIX0", READ)
       rejpix1_col = column(exposureTable, "REJPIX1", READ)
       rejpix0 => int32Data( rejpix0_col )
       rejpix1 => int32Data( rejpix1_col )

       nuplinked = 0
       do inode=0,1                    !loop over nodes
          call CAL_setState(ccdnodeid=inode)
          call CAL_getBadPixelList(rawX, rawY, badtype, yextent, uplinked)

          if ( size(rawX) > 0 ) then
             do i=0,ubound(rawx,1)
                if ( uplinked(i) ) nuplinked(inode+1) = nuplinked(inode+1) + yextent(i)
             end do
          end if
       end do
                                !all bins that were read out (total active surface)
       allbins = int16Attribute(pixelsTable,"WINDOWDX") * int16Attribute(pixelsTable,"WINDOWDY")

                                !telemetered fraction of active surface
       fracexp0 = 1. - real(rejpix0) / real( allbins - nuplinked(1) )
       fracexp1 = 1. - real(rejpix1) / real( allbins - nuplinked(2) )

       call release( rejpix0_col )
       call release( rejpix1_col )

    else
       fracexp0 = 1.
       fracexp1 = 1.
    end if

    call release( fracexp0_col )
    call release( fracexp1_col )
  end subroutine addFracExposure


  !---------------------------------------------------------------------
  subroutine cleanupExposureTable
  !---------------------------------------------------------------------
    !delete columns that should not be propagated

    if ( notHTR ) then
       call deleteColumn(exposureTable, 'SEQINDEX')
       call deleteColumn(exposureTable, 'REJPIX0')
       call deleteColumn(exposureTable, 'REJPIX1')
       call deleteColumn(exposureTable, 'EOSCOARS')
       call deleteColumn(exposureTable, 'EOSFINE' )
       call deleteColumn(exposureTable, 'NDPP')

    else
       call deleteColumn(exposureTable, 'NLOSTEVT')
       call deleteColumn(exposureTable, 'ABORTFLG')
    end if

    call release(exposureTable)
  end subroutine cleanupExposureTable


  !------------------------------------------------------------
  subroutine calculateHTRFrameTimes (set)
    !------------------------------------------------------------
    type(DataSetT) :: set       !a dummy MODIFY-able set
    type(TableT) :: auxTable    !the AUX table
    type(ColumnT) :: auxframe_col, auxccdid_col!, eoscoars_col, eosfine_col
    integer(kind=int32), dimension(:), pointer :: auxframe => null()!, eoscoars, eosfine
    integer(kind=int8),  dimension(:), pointer :: auxccdid => null()

    type(tableT) :: auxTable2   !the AUX2 table
    type(ColumnT) :: seqCCDIDList_col
    integer(kind=int8), dimension(:), pointer :: seqCCDIDList => null()

                                !CCD event lists
    character(len=6) :: evlistname
    type(DataSetT), dimension(9) :: ccdevset
    type(TableT), dimension(9) :: evTable, TmpEvTable
    logical, dimension(9) :: ccdevlistopen
    type(ColumnT) :: evframe_col, deltatime_col, col
    integer(kind=int32), dimension(:), pointer :: evframe => null(), deltatime => null()

    integer :: instrument, seqLength, i, evrow, lastRow(9), n
    character(len=8) :: TmpTabName

    call message(appMsg, Verbose, "Starting calculateHTRFrameTimes")

                                !need to access AUX table again
    call CAL_getState(instrument=instrument)
    select case (instId(instrument)) !table names depend on instrument
    case ('RGS1')
       auxTable = table(auxSet, 'R1AUX1', READ)
    case ('RGS2')
       auxTable = table(auxSet, 'R2AUX1', READ)
    case default
       call error ('instIdNotFoundExposure','instrument string '//instId(instrument))
    end select

    auxframe_col = column(auxTable, FRAME_COLUMN_NAME, READ)
    auxframe    => int32Data( auxframe_col )
    auxccdid_col = column(auxTable, CCDID_COLUMN_NAME, READ)
    auxccdid    => int8data( auxccdid_col )

    select case (instId(instrument)) !table names depend on instrument
    case ('RGS1')
       auxTable2 = table(auxSet, 'R1AUX2', READ)
       evlistname = 'R1HTE1'
    case ('RGS2')
       auxTable2 = table(auxSet, 'R2AUX2', READ)
       evlistname = 'R2HTE1'
    end select
                                !get the readout sequence from AUX2
    seqCCDIDList_col = column(auxTable2, 'CCDID', READ)
    seqCCDIDList => int8Data (seqCCDIDList_col)
    seqLength = count(seqCCDIDList /= 0) !total length of readout sequence

                                !open CCD event lists to read delta_times
    ccdevlistopen = .false.
    OpenEvLists: do i=1,seqLength
                                !same CCD may be more than once in sequence
       if ( .not. ccdevlistopen(seqCCDIDList(i-1)) ) then
          ccdevset(i) = dataset( OAL_selectSet(int(seqCCDIDList(i-1))) , READ)
          call message(appMsg, Verbose, "Opened event list "//trim(OAL_selectSet(int(seqCCDIDList(i-1)))))
          evTable(seqCCDIDList(i-1)) = table( ccdevset(i), evlistname )

                                !need to convert RAWY from unsigned16 to int32
          col =  Column(evTable(seqCCDIDList(i-1)), "RAWY", READ)
          write(TmpTabName,'(a5,i1)') "TMPDT",i
          TmpEvTable(seqCCDIDList(i-1)) = addTable( set, TmpTabName, numberOfRows(evTable(seqCCDIDList(i-1))))
          deltatime_col = addColumn(TmpEvTable(seqCCDIDList(i-1)),"DELTA_T", integer32)
          call setData( deltatime_col, col )
!          deltatime => int32Data( deltatime_col )

          call release( col )
          call release( deltatime_col )

          ccdevlistopen(seqCCDIDList(i-1)) = .true.
       end if
    end do OpenEvLists
                                !create an array with all delta_T
    allocate(alldeltaT(0:size(auxframe)-1))
    alldeltaT = -1

    avgDTime = 0
    lastRow = 0

!?!    eoscoars_col = column(auxTable, 'EOSCOARS', READ)
!?!    eoscoars => int32Data( eoscoars_col )

    call message(appMsg, Verbose, "Creating AllDeltaT ...")
    CreateAllDeltaT: do i=0,size(auxframe)-1
!       if ( eoscoars(i) > 0 ) print*,'Row=',i,'EOSCOARS=',eoscoars(i) 
                                !missing frame in AUX file
       if ( i > 0 ) then
          if ( auxframe(i)-auxframe(i-1) /= 1 ) cycle CreateAllDeltaT
       end if

       if ( auxccdid(i) < 1 .or. auxccdid(i) > 9 .or. .not. any(auxccdid(i)==seqCCDIDList) ) then
          write(errstr,*) 'Row',i+1,'of AUX file (FRAME=', auxframe(i),') has invalid CCDID: CCDID=', auxccdid(i)
          call warning('HTRAUXCCDID')
          cycle CreateAllDeltaT
       end if

       deltatime_col = column( TmpEvTable(auxccdid(i)), "DELTA_T", READ)
       evframe_col   = column( evTable(auxccdid(i)), FRAME_COLUMN_NAME, READ)
       deltatime => int32Data( deltatime_col )
       evframe   => int32Data( evframe_col )

       FindRow: do evrow=lastRow(auxccdid(i)),ubound(evframe,1)
          if ( evframe(evrow) == auxframe(i) ) then
             alldeltaT(i) = deltatime(evrow) / 65536.
             lastRow(auxccdid(i)) = evrow

                                !prepare averaging of delta_T's
             avgDTime(auxccdid(i)) = avgDTime(auxccdid(i)) + alldeltaT(i)
             exit FindRow

          else if ( evframe(evrow) > auxframe(i) ) then !entry is missing
             exit FindRow
          end if
       end do FindRow

       call release( deltatime_col )
       call release( evframe_col )

       if ( alldeltaT(i) > 0.5 ) then !check error in instrument
          write(errstr,*) 'DeltaT>0.5 found. DeltaT=',alldeltaT(i),'sec. Frame=',&
               auxframe(i),'CCDID=',auxccdid(i),'in row', evrow,'of event list.'
          call warning('HTRDataError')
       end if
    end do CreateAllDeltaT

    call message(appMsg, Verbose, "Finished of AllDeltaT generation.")

    CloseEvLists: do i=1,seqLength !rlease CCD event list sets
       if ( ccdevlistopen(seqCCDIDList(i-1)) ) then
          call release( evTable(seqCCDIDList(i-1)) )

          call release( TmpEvTable(seqCCDIDList(i-1)) ) !remove temporary table
          write(TmpTabName,'(a5,i1)') "TMPDT",i
          call deleteBlock( set, TmpTabName )

          call release( ccdevset(i) )

          ccdevlistopen(seqCCDIDList(i-1)) = .false.
       end if
    end do CloseEvLists

    call release( seqCCDIDList_col ) !components of AUX2 table
    call release( auxTable2 )
                                    !create average delta_T's per CCD
    do i=1,9
       n = count( alldeltaT > 0 .and. auxccdid==i )
       if ( n > 0 ) avgDTime(i) = avgDTime(i) / n
    end do

!    print*, "CCDID's=0 at rows:",pack( auxccdid, auxccdid<1 ), count( auxccdid<1 )
    n = count( alldeltaT < 0 )
    if ( n > 0 ) then
       write(errstr,*) n, "Frames filled with average delta_T's"
       call warning ('missingDeltaT')
       where( alldeltaT < 0 .and. auxccdid > 0 ) alldeltaT = avgDTime( auxccdid )
    end if

    call release( auxframe_col ) !from the AUX table
    call release( auxccdid_col )
    call release( auxTable )

    call message(appMsg, Verbose, "Finished calculateHTRFrameTimes")
  end subroutine calculateHTRFrameTimes


  !------------------------------------------------------------
  subroutine fillHTRFrameTimes
    !------------------------------------------------------------
                                !EXPOSURE table
    type(ColumnT) :: frame_col, timedel_col, time_col, flag_col
    integer(kind=int32), dimension(:), pointer :: frame => null(), flag => null()
    real, dimension(:), pointer :: timedel => null()
    real(double), dimension(:), pointer :: time => null()

    type(TableT) :: auxTable    !the AUX table
    type(ColumnT) :: auxframe_col, auxccdid_col
    integer(kind=int32), dimension(:), pointer :: auxframe => null()
    integer(kind=int8),  dimension(:), pointer :: auxccdid => null()

    integer :: instrument, theccdid, exprow, i
    real(double) :: expStartTime, timenow, lasttime

    call message(appMsg, Verbose, "Starting fillHTRFrameTimes")

                                !create time columns to EXPOSURE table
    time_col = addColumn(exposureTable,TIME_COLUMN_NAME,real64,units='s',comment='seconds since MJDREF')
    time => real64Data(time_col)
    timedel_col = addColumn(exposureTable,'TIMEDEL',real32,units='s',comment='Frame integration time')
    timedel => real32Data(timedel_col)
    timedel = 0

    frame_col = column(exposureTable, FRAME_COLUMN_NAME, READ)
    frame => int32Data( frame_col )

                                !need to access AUX table again
    call CAL_getState(instrument=instrument,ccdChipId=theccdid, &
         exposureStartTime=expStartTime )
    select case (instId(instrument)) !table names depend on instrument
    case ('RGS1')
       auxTable = table(auxSet, 'R1AUX1', READ)
    case ('RGS2')
       auxTable = table(auxSet, 'R2AUX1', READ)
    case default
       call error ('instIdNotFoundExposure','instrument string '//instId(instrument))
    end select

    auxframe_col = column(auxTable, FRAME_COLUMN_NAME, READ)
    auxframe    => int32Data( auxframe_col )
    auxccdid_col = column(auxTable, CCDID_COLUMN_NAME, READ)
    auxccdid    => int8data( auxccdid_col )

                                !translate start of exposure to XMM ref time
    timenow = toTimeTag( stringAttribute(auxTable,"DATE-OBS") ) &
         - toTimeTag("1998-01-01-T00:00:00") !XMM reference time

                                !loop over all frames in AUX file and add delta_times
    exprow = 0
    lasttime = 0
    AddTimes: do i=0,size(alldeltaT)-1

       timenow = timenow + allDeltaT(i)   !running sum of all times

       if ( auxccdid(i) == theccdid .and. auxframe(i) == frame(exprow) ) then
                                !event time & frame integration time
          timedel(exprow) = timenow - lasttime
          time(exprow) = timenow - timedel(exprow) /2.
          lasttime = timenow

          if ( exprow == ubound(frame,1) ) then
             exit AddTimes
          else
             exprow = exprow + 1
          end if
       end if
    end do AddTimes

    call release( auxframe_col ) !from the AUX table
    call release( auxccdid_col )
    call release( auxTable )

                                !check lower limit on frame times
    flag_col = Column(exposureTable, EVATT_COLUMN_NAME)
    flag => int32Data(flag_col)
    where ( timedel < realParameter("minexposuretime") )&
      flag = EVATT_IN_BAD_FRAME

    if ( verbosity(appMsg) >= Silent) then
       write(errstr,*) count( timedel < realParameter("minexposuretime") ), &
            "frames flagged bad based on minexposuretime"
       call message(appMsg,Silent)
    end if

    if (all(timedel < realParameter("minexposuretime"))) then
      write(errstr,'(a, i2.2)') "All frames flagged as EVATT_IN_BAD_FRAME &
      &because TIMEDEL shorter than --minexposuretime: &
      &therefore no GTI for ccd ", theccdid
      call warning('htrAllFramesTooShort')
    end if

    call release( frame_col )   !components of the EXPOSURE table
    call release( timedel_col )
    call release( time_col )
    call release( flag_col )

    call message(appMsg, Verbose, "Finished fillHTRFrameTimes")
  end subroutine fillHTRFrameTimes
end module rgsframes_mod
